﻿using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http.Features;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using Netmash.Security.Authentication.Tuitio.Abstractions;
using Netmash.Security.Authentication.Tuitio.Services;
using System;
using System.Net.Http.Headers;
using System.Text.Encodings.Web;
using System.Threading.Tasks;
using Tuitio.PublishedLanguage.Dto;
using Tuitio.Wrapper.Services;
using c = Netmash.Security.Authentication.Tuitio.Constants;

namespace Netmash.Security.Authentication.Tuitio
{
    public class TuitioAuthenticationHandler : AuthenticationHandler<AuthenticationSchemeOptions>
    {
        private readonly ITuitioService _tuitioService;
        private readonly IAuthenticationOptions _authenticationOptions;
        private readonly ILogger<TuitioAuthenticationHandler> _logger;

        public TuitioAuthenticationHandler(IOptionsMonitor<AuthenticationSchemeOptions> options, ILoggerFactory loggerFactory, UrlEncoder encoder, ISystemClock clock, ITuitioService tuitioService, IAuthenticationOptions authenticationOptions, ILogger<TuitioAuthenticationHandler> logger)
            : base(options, loggerFactory, encoder, clock)
        {
            _tuitioService = tuitioService;
            _authenticationOptions = authenticationOptions;
            _logger = logger;
        }

        private AuthenticationTicketService TicketService => new(Scheme.Name, _logger);

        protected override async Task<AuthenticateResult> HandleAuthenticateAsync()
        {
            // if the method is marked with [AllowAnonymous], the handler will skip token validation.
            var endpointFeature = Context.Features.Get<IEndpointFeature>();
            var endpoint = endpointFeature?.Endpoint;
            if (endpoint?.Metadata.GetMetadata<IAllowAnonymous>() != null)
            {
                return AuthenticateResult.NoResult();
            }

            var token = GetAuthorizationToken();
            if (token != null)
            {
                Envelope<AuthorizationResult> authorizationEnvelope;
                try
                {
                    authorizationEnvelope = await _tuitioService.Authorize(token);
                }
                catch (Exception e)
                {
                    _logger.LogError(e, "Tuitio authorization failed");
                    return AuthenticateResult.Fail("UNAUTHORIZED");
                }

                if (!string.IsNullOrEmpty(authorizationEnvelope.Error))
                    return AuthenticateResult.Fail(authorizationEnvelope.Error);

                var ticket = TicketService.GetAuthenticationTicket(authorizationEnvelope.Result);
                return AuthenticateResult.Success(ticket);
            }

            var authenticateAsGuest = _authenticationOptions.AuthenticateAsGuest?.Invoke(Request) ?? false;
            if (authenticateAsGuest)
            {
                var guestTicket = TicketService.GetGuestAuthenticationTicket(_authenticationOptions.GuestUserId, _authenticationOptions.GuestUserName);
                return AuthenticateResult.Success(guestTicket);
            }

            return AuthenticateResult.Fail("AUTHORIZATION_HEADER_IS_MISSING");
        }

        private string GetAuthorizationToken()
        {
            if (Request.Headers.ContainsKey("Authorization"))
            {
                var authorizationHeader = AuthenticationHeaderValue.Parse(Request.Headers["Authorization"]);
                var token = authorizationHeader.Parameter;
                return token;
            }

            if (_authenticationOptions.AcceptTokenFromQuery
                && Request.Query.Count > 0
                && Request.Query.ContainsKey(c.QueryParams.Token))
            {
                var token = Request.Query[c.QueryParams.Token];
                return token.ToString();
            }

            return null;
        }
    }
}
